# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any, Dict, List, Optional

import libyang

from _sysrepo import lib


# ------------------------------------------------------------------------------
class Change:

    __slots__ = ("xpath",)

    def __init__(self, xpath: str):
        self.xpath = xpath

    class Skip(Exception):
        """
        Risen when a change should be skipped.
        """

        pass

    @staticmethod
    def parse(
        operation: int,
        node: libyang.DNode,
        prev_val: str,
        prev_list: str,
        prev_dflt: bool,
        include_implicit_defaults: bool = True,
    ) -> "Change":
        """
        Parse an operation code and values from libsysrepo.so and return a Change
        subclass instance.

        Meaning of parameters varies based on the operation:

        SR_OP_CREATED
            node is the created node, for user-ordered lists either prev_value or
            prev_list is always set with meaning similar to SR_OP_MOVED.
        SR_OP_MODIFIED
            node is the modified node, prev_value is set to the previous value of the
            leaf, prev_dflt is set if the previous leaf value was the default.
        SR_OP_DELETED
            node is the deleted node.
        SR_OP_MOVED
            node is the moved (leaf-)list instance, for user-ordered lists either
            prev_value (leaf-list) or prev_list (list) is set to the preceding instance
            unless the node is the first, when they are set to "" (empty string).

        :arg operation:
            The operation code.
        :arg node:
            Affected data node always with all parents, depends on the operation.
        :arg prev_val:
            Previous value, depends on the operation.
        :arg prev_list:
            Previous list keys predicate (`[key1="val1"][key2="val2"]...`), depends on
            the operation.
        :arg prev_dflt:
            Previous value default flag, depends on the operation.
        :arg include_implicit_defaults:
            Include implicit default values into the data dictionaries.
        """
        if operation == lib.SR_OP_CREATED:
            if not node.should_print(
                include_implicit_defaults=include_implicit_defaults
            ):
                raise Change.Skip()
            return ChangeCreated(
                node.path(),
                _node_value(node, include_implicit_defaults),
                after=_after_key(node, prev_val, prev_list),
            )
        if operation == lib.SR_OP_MODIFIED:
            return ChangeModified(
                node.path(),
                _node_value(node, include_implicit_defaults),
                prev_val=prev_val,
                prev_dflt=prev_dflt,
            )
        if operation == lib.SR_OP_DELETED:
            return ChangeDeleted(node.path())
        if operation == lib.SR_OP_MOVED:
            return ChangeMoved(
                node.path(), after=_after_key(node, prev_val, prev_list),
            )
        raise ValueError("unknown change operation: %s" % operation)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.xpath == other.xpath

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    __hash__ = None  # not hashable

    def __repr__(self) -> str:
        return "%s(%s)" % (type(self).__name__, self)

    def __str__(self) -> str:
        return self.xpath


# ------------------------------------------------------------------------------
class ChangeCreated(Change):

    __slots__ = Change.__slots__ + ("value", "after")

    def __init__(self, xpath: str, value: Any, after: Optional[str] = None):
        super().__init__(xpath)
        self.value = value
        self.after = after

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and other.after == self.after
            and other.value == self.value
        )

    def __str__(self) -> str:
        if self.after == ():
            return "%s: %r: FIRST" % (self.xpath, self.value)
        if self.after is not None:
            return "%s: %r: AFTER %r" % (self.xpath, self.value, self.after)
        return "%s: %r" % (self.xpath, self.value)


# ------------------------------------------------------------------------------
class ChangeModified(Change):

    __slots__ = Change.__slots__ + ("value", "prev_val", "prev_dflt")

    def __init__(self, xpath: str, value: Any, prev_val: str, prev_dflt: bool = False):
        super().__init__(xpath)
        self.value = value
        self.prev_val = prev_val
        self.prev_dflt = prev_dflt

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and other.prev_dflt == self.prev_dflt
            and other.prev_val == self.prev_val
            and other.value == self.value
        )

    def __str__(self) -> str:
        return "%s: %r -> %r" % (self.xpath, self.prev_val, self.value)


# ------------------------------------------------------------------------------
class ChangeDeleted(Change):
    pass


# ------------------------------------------------------------------------------
class ChangeMoved(Change):

    __slots__ = Change.__slots__ + ("after",)

    def __init__(self, xpath: str, after: str):
        super().__init__(xpath)
        self.after = after

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and other.after == self.after

    def __str__(self) -> str:
        if self.after:
            where = "AFTER %r" % (self.after,)
        else:
            where = "FIRST"
        return "%s: %s" % (self.xpath, where)


# -------------------------------------------------------------------------------------
def update_config_cache(conf: Dict, changes: List[Change]) -> None:
    """
    Maintain a configuration dict from a list of Change objects.

    This function is intended to be used in module change callbacks if they want to
    preserve a full view of all the configuration without asking sysrepo everytime the
    callback is invoked.

    :arg conf:
        The cached config dict to update.
    :arg changes:
        The list of changes passed to module change callbacks.
    """
    for c in changes:
        if isinstance(c, ChangeCreated):
            libyang.xpath_set(conf, c.xpath, c.value, after=c.after)
        elif isinstance(c, ChangeModified):
            libyang.xpath_set(conf, c.xpath, c.value)
        elif isinstance(c, ChangeMoved):
            libyang.xpath_move(conf, c.xpath, c.after)
        elif isinstance(c, ChangeDeleted):
            libyang.xpath_del(conf, c.xpath)


# -------------------------------------------------------------------------------------
def _node_value(node: libyang.DNode, include_implicit_defaults: bool = True) -> Any:
    """
    Extract a python value from a libyang.DNode.
    """
    if isinstance(node, (libyang.DLeaf, libyang.DLeafList)):
        return node.value()
    dic = node.print_dict(
        absolute=False, include_implicit_defaults=include_implicit_defaults
    )
    if not dic:
        return dic
    dic = next(iter(dic.values()))  # trim first level of dict with only key name
    if isinstance(node, libyang.DList):
        dic = next(iter(dic))  # only preserve the list element dict
    return dic


# -------------------------------------------------------------------------------------
def _after_key(
    node: libyang.DNode, prev_val: Optional[str], prev_list: Optional[str]
) -> Optional[str]:
    """
    Get the proper `after` key value based on the node type.
    """
    if isinstance(node, libyang.DLeafList):
        return prev_val
    if isinstance(node, libyang.DList):
        return prev_list
    return None



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import asyncio
import functools
import logging
from typing import Any, Callable

from libyang.data import DNode

from _sysrepo import ffi, lib
from .errors import SysrepoError, check_call
from .util import c2str, is_async_func


LOG = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
class Subscription:
    """
    Python representation of `sr_subscription_ctx_t *`.

    .. attention::

        Do not instantiate this class manually, use `SysrepoSession.subscribe_*`.
    """

    def __init__(
        self,
        callback: Callable,
        private_data: Any = None,
        asyncio_register: bool = False,
        strict: bool = False,
        include_implicit_defaults: bool = True,
    ):
        """
        :arg callback:
            The python callback function or coroutine function used when subscribing.
        :arg private_data:
            Opaque data used when subscribing, will be passed to callback.
        :arg asyncio_register:
            Add this subscription event pipe into asyncio event loop monitored file
            descriptors. When the event pipe becomes readable, call process_events().
        :arg strict:
            If True, reject data with no schema definition from rpc output parameters
            and operational data callbacks. Otherwise, ignore unknown data and log a
            warning message.
        :arg include_implicit_defaults:
            If True, include implicit default nodes into Change objects passed to module
            change callbacks and into input parameters passed to RPC/action callbacks.
        """
        if is_async_func(callback) and not asyncio_register:
            raise ValueError(
                "%s is an async function, asyncio_register is mandatory" % callback
            )
        self.callback = callback
        self.private_data = private_data
        self.asyncio_register = asyncio_register
        self.strict = strict
        self.include_implicit_defaults = include_implicit_defaults
        if asyncio_register:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = None
        self.tasks = {}
        self.cdata = None
        self.fd = -1
        self.handle = ffi.new_handle(self)

    def init(self, cdata) -> None:
        """
        Initialization of this object is not complete after calling __init__. The
        sr_subscription_ctx_t object is allocated by sysrepo when calling one of
        sr_*_subscribe functions and we need to pass self.handle to these functions so
        that this subscription can be forwarded to C callbacks.

        Subscription.init() is called just after sr_*_subscribe functions to complete
        initialization. See SysrepoSession.subscribe_* functions for more details.

        if self.asyncio_register is  True, add this subscription event pipe to the
        monitored file descriptors for reading in asyncio event loop.

        :arg "sr_subscription_ctx_t *" cdata:
            The subscription pointer allocated by sysrepo.
        """
        if self.cdata is not None:
            raise RuntimeError("init was already called once")
        self.cdata = cdata
        if self.asyncio_register:
            self.fd = self.get_fd()
            self.loop.add_reader(self.fd, self.process_events)

    def get_fd(self) -> int:
        """
        Get the event pipe of a subscription. Event pipe can be used in `select()`,
        `poll()`, or similar functions to listen for new events. It will then be ready
        for reading.
        """
        fd_p = ffi.new("int *")
        check_call(lib.sr_get_event_pipe, self.cdata, fd_p)
        return fd_p[0]

    def unsubscribe(self) -> None:
        """
        Unsubscribes from a subscription acquired by any of sr_*_subscribe calls and
        releases all subscription-related data.

        Removes self.fd from asyncio event loop monitored file descriptors.
        """
        if self.cdata is None:
            return
        if self.asyncio_register and self.fd != -1:
            self.loop.remove_reader(self.fd)
        try:
            check_call(lib.sr_unsubscribe, self.cdata)
        finally:
            self.cdata = None
        for t in list(self.tasks.values()):
            t.cancel()
        self.tasks.clear()

    def process_events(self) -> None:
        """
        Called when self.fd becomes readable.
        """
        check_call(lib.sr_process_events, self.cdata, ffi.NULL, ffi.NULL)

    def task_done(self, task_id: Any, event: str, task: asyncio.Task) -> None:
        """
        Called when self.callback is an async function/method and it has finished. This
        calls self.process_events() so that the C callback is invoked again with the
        same arguments (request_id, event) and we can return the actual result.
        """
        if task.cancelled():
            self.tasks.pop(task_id, None)
            return
        try:
            if event in ("update", "change", "rpc", "oper"):
                # The task result will be evaluated in the C callback.
                # It will return the result to sysrepo.
                self.process_events()
            else:
                # Sysrepo does not care about the result of the callback.
                # This will raise the exception here if any occured in the task
                # and will be logged (i.e. not lost).
                self.tasks.pop(task_id, None)
                task.result()
        except Exception:
            LOG.exception("failure in task: %r", task)


# ------------------------------------------------------------------------------
EVENT_NAMES = {
    lib.SR_EV_UPDATE: "update",
    lib.SR_EV_CHANGE: "change",
    lib.SR_EV_DONE: "done",
    lib.SR_EV_ABORT: "abort",
    lib.SR_EV_ENABLED: "enabled",
    lib.SR_EV_RPC: "rpc",
}


# ------------------------------------------------------------------------------
@ffi.def_extern(name="srpy_module_change_cb")
def module_change_callback(session, module, xpath, event, req_id, priv):
    """
    Callback to be called on the event of changing datastore content of the specified
    module.

    This python function mapped to the C srpy_module_change_cb function. When the C
    srpy_module_change_cb function is called by libsysrepo.so, this function is called
    with the same arguments.

    :arg "sr_session_ctx_t *" session:
        Implicit session (do not stop) with information about the changed data.
    :arg "const char *" module:
        Name of the module where the change has occurred.
    :arg "const char *" xpath:
        XPath used when subscribing, NULL if the whole module was subscribed to.
    :arg "sr_event_t" event:
        Type of the callback event that has occurred.
    :arg "uint32_t" req_id:
        Request ID unique for the specific module_name. Connected events for one request
        (SR_EV_CHANGE and SR_EV_DONE, for example) have the same request ID.
    :arg "void *" priv:
        Private context opaque to sysrepo. Contains a CFFI handle to the Subscription
        python object.

    :returns:
        User error code (sr_error_t).
    :raises:
        IMPORTANT: This function *CANNOT* raise any exception. The C callstack does not
        handle that well and when it happens the outcome is undetermined. Make sure to
        catch all errors and log them so they are not lost.
    """
    try:
        # convert C arguments to python objects.
        from .session import SysrepoSession  # circular import

        session = SysrepoSession(session, True)
        module = c2str(module)
        xpath = c2str(xpath)
        root_xpath = ("/%s:*" % module) if xpath is None else xpath
        subscription = ffi.from_handle(priv)
        callback = subscription.callback
        private_data = subscription.private_data
        event_name = EVENT_NAMES[event]

        if is_async_func(callback):
            task_id = (event, req_id)

            if task_id not in subscription.tasks:
                # ATTENTION: the implicit session passed as argument will be
                # freed when this function returns. The async callback must NOT
                # keep a reference on it as it will be invalid. Changes must be
                # gathered now.
                changes = list(
                    session.get_changes(
                        root_xpath + "//.",
                        include_implicit_defaults=subscription.include_implicit_defaults,
                    )
                )
                task = subscription.loop.create_task(
                    callback(event_name, req_id, changes, private_data)
                )
                task.add_done_callback(
                    functools.partial(subscription.task_done, task_id, event_name)
                )
                subscription.tasks[task_id] = task
                if event not in (lib.SR_EV_UPDATE, lib.SR_EV_CHANGE):
                    # Return immediately, process_events will not be called in
                    # subscription.task_done. Sysrepo does not care about the
                    # result of the operation.
                    return lib.SR_ERR_OK

            task = subscription.tasks[task_id]

            if not task.done():
                return lib.SR_ERR_CALLBACK_SHELVE

            del subscription.tasks[task_id]
            task.result()  # raise error if any

        else:
            changes = list(
                session.get_changes(
                    root_xpath + "//.",
                    include_implicit_defaults=subscription.include_implicit_defaults,
                )
            )
            callback(event_name, req_id, changes, private_data)

        return lib.SR_ERR_OK

    except SysrepoError as e:
        if (
            event in (lib.SR_EV_UPDATE, lib.SR_EV_CHANGE)
            and e.msg
            and isinstance(session, SysrepoSession)
            and isinstance(xpath, str)
        ):
            session.set_error(xpath, e.msg)
        return e.rc

    except BaseException as e:
        # ATTENTION: catch all exceptions!
        # including KeyboardInterrupt, CancelledError, etc.
        # We are in a C callback, we cannot let any error pass
        LOG.exception("%r callback failed", locals().get("callback", priv))
        if (
            event in (lib.SR_EV_UPDATE, lib.SR_EV_CHANGE)
            and isinstance(session, SysrepoSession)
            and isinstance(xpath, str)
        ):
            session.set_error(xpath, str(e))
        return lib.SR_ERR_CALLBACK_FAILED


# ------------------------------------------------------------------------------
@ffi.def_extern(name="srpy_oper_data_cb")
def oper_data_callback(session, module, xpath, req_xpath, req_id, parent, priv):
    """
    Callback to be called when operational data at the selected xpath are requested.

    :arg "sr_session_ctx_t *" session:
        Implicit session (do not stop).
    :arg "const char *" module:
        Name of the affected module.
    :arg "const char *" xpath:
        XPath identifying the subtree that is supposed to be provided, same as the one
        used for the subscription.
    :arg "const char *" req_xpath:
        XPath as requested by a client. Can be NULL.
    :arg "uint32_t" req_id:
        Request ID unique for the specific module name.
    :arg "struct lyd_node **" parent:
        Pointer to an existing parent of the requested nodes. Is NULL for top-level
        nodes. Callback is supposed to append the requested nodes to this data subtree
        and return either the original parent or a top-level node.
    :arg "void *" priv:
        Private context opaque to sysrepo. Contains a CFFI handle to the Subscription
        python object.

    :returns:
        User error code (sr_error_t).
    :raises:
        IMPORTANT: This function *CANNOT* raise any exception. The C callstack does not
        handle that well and when it happens the outcome is undetermined. Make sure to
        catch all errors and log them so they are not lost.
    """
    try:
        # convert C arguments to python objects.
        from .session import SysrepoSession  # circular import

        session = SysrepoSession(session, True)
        module = c2str(module)
        xpath = c2str(xpath)
        req_xpath = c2str(req_xpath)
        subscription = ffi.from_handle(priv)
        callback = subscription.callback
        private_data = subscription.private_data

        if is_async_func(callback):
            task_id = req_id

            if task_id not in subscription.tasks:
                task = subscription.loop.create_task(callback(req_xpath, private_data))
                task.add_done_callback(
                    functools.partial(subscription.task_done, task_id, "oper")
                )
                subscription.tasks[task_id] = task

            task = subscription.tasks[task_id]

            if not task.done():
                return lib.SR_ERR_CALLBACK_SHELVE

            del subscription.tasks[task_id]

            oper_data = task.result()

        else:
            oper_data = callback(req_xpath, private_data)

        if isinstance(oper_data, dict):
            # convert oper_data to a libyang.DNode object
            ly_ctx = session.get_ly_ctx()
            dnode = ly_ctx.get_module(module).parse_data_dict(
                oper_data, data=True, strict=subscription.strict, validate=False
            )
            if dnode is not None:
                if parent[0]:
                    root = DNode.new(ly_ctx, parent[0]).root()
                    root.merge(dnode, destruct=True)
                else:
                    # The FFI bindings of libyang and sysrepo are different.
                    # Casting is required.
                    parent[0] = ffi.cast("struct lyd_node *", dnode.cdata)
        elif oper_data is not None:
            raise TypeError(
                "bad return type from %s (expected dict or None)" % callback
            )

        return lib.SR_ERR_OK

    except SysrepoError as e:
        if e.msg and isinstance(session, SysrepoSession) and isinstance(xpath, str):
            session.set_error(xpath, e.msg)
        return e.rc

    except BaseException as e:
        # ATTENTION: catch all exceptions!
        # including KeyboardInterrupt, CancelledError, etc.
        # We are in a C callback, we cannot let any error pass
        LOG.exception("%r callback failed", locals().get("callback", priv))
        if isinstance(session, SysrepoSession) and isinstance(xpath, str):
            session.set_error(xpath, str(e))
        return lib.SR_ERR_CALLBACK_FAILED


# ------------------------------------------------------------------------------
@ffi.def_extern(name="srpy_rpc_tree_cb")
def rpc_callback(session, xpath, input_node, event, req_id, output_node, priv):
    """
    Callback to be called for the delivery of an RPC/action.

    :arg "sr_session_ctx_t *" session:
        Implicit session (do not stop).
    :arg "const char *" xpath:
        Simple operation path identifying the RPC/action.
    :arg "const struct lyd_node *" input_node:
        Data tree of input parameters.
    :arg "sr_event_t" event:
        Type of the callback event that has occurred.
    :arg "uint32_t" req_id:
        Request ID unique for the specific xpath.
    :arg "struct lyd_node *" output_node:
        Data tree of output parameters. Should be allocated on heap, will be freed by
        sysrepo after sending of the RPC response.
    :arg "void *" priv:
        Private context opaque to sysrepo. Contains a CFFI handle to the Subscription
        python object.

    :returns:
        User error code (sr_error_t).
    :raises:
        IMPORTANT: This function *CANNOT* raise any exception. The C callstack does not
        handle that well and when it happens the outcome is undetermined. Make sure to
        catch all errors and log them so they are not lost.
    """
    try:
        # convert C arguments to python objects.
        from .session import SysrepoSession  # circular import

        session = SysrepoSession(session, True)
        subscription = ffi.from_handle(priv)
        callback = subscription.callback
        private_data = subscription.private_data
        event_name = EVENT_NAMES[event]
        ly_ctx = session.get_ly_ctx()
        rpc_input = DNode.new(ly_ctx, input_node)
        xpath = rpc_input.path()
        # strip all parents, only preserve the input tree
        input_dict = next(
            iter(
                rpc_input.print_dict(
                    include_implicit_defaults=subscription.include_implicit_defaults,
                    absolute=False,
                ).values()
            )
        )

        if is_async_func(callback):
            task_id = (event, req_id)

            if task_id not in subscription.tasks:
                task = subscription.loop.create_task(
                    callback(xpath, input_dict, event_name, private_data)
                )
                task.add_done_callback(
                    functools.partial(subscription.task_done, task_id, event_name)
                )
                subscription.tasks[task_id] = task

            task = subscription.tasks[task_id]

            if not task.done():
                return lib.SR_ERR_CALLBACK_SHELVE

            del subscription.tasks[task_id]

            output_dict = task.result()

        else:
            output_dict = callback(xpath, input_dict, event_name, private_data)

        if event != lib.SR_EV_RPC:
            # May happen when there are multiple callback registered for the
            # same RPC. If one of the callbacks has failed, the other ones will
            # be called with SR_EV_ABORT. In that case, abort early and do
            # not return the RPC output data to sysrepo.
            return lib.SR_ERR_OK

        if isinstance(output_dict, dict):
            # update output_node with contents of output_dict
            DNode.new(ly_ctx, output_node).merge_data_dict(
                output_dict, rpcreply=True, strict=subscription.strict, validate=False
            )
        elif output_dict is not None:
            raise TypeError(
                "bad return type from %s (expected dict or None)" % callback
            )

        return lib.SR_ERR_OK

    except SysrepoError as e:
        if e.msg and isinstance(session, SysrepoSession) and isinstance(xpath, str):
            session.set_error(xpath, e.msg)
        return e.rc

    except BaseException as e:
        # ATTENTION: catch all exceptions!
        # including KeyboardInterrupt, CancelledError, etc.
        # We are in a C callback, we cannot let any error pass
        LOG.exception("%r callback failed", locals().get("callback", priv))
        if isinstance(session, SysrepoSession) and isinstance(xpath, str):
            session.set_error(xpath, str(e))
        return lib.SR_ERR_CALLBACK_FAILED



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import logging
from typing import Optional, Sequence

import libyang

from _sysrepo import ffi, lib
from .errors import SysrepoInternalError, check_call
from .session import SysrepoSession, datastore_value
from .util import str2c


LOG = logging.getLogger(__name__)


# ------------------------------------------------------------------------------
class SysrepoConnection:
    """
    Create a connection to the sysrepo datastore. If possible (no other connections
    exist), also apply any scheduled changes.

    Do not use `os.fork` after creating a connection. Sysrepo internally stores PID of
    every created connection and this way a mismatch of PID and connection is created.

    The created object can be used as a context manager and will be automatically
    "closed" on exit::

        with sysrepo.SysrepoConnection() as conn:
            # to stuff with conn
        # conn.disconnect() has been called whatever happens
    """

    __slots__ = ("cdata",)

    def __init__(
        self,
        cache_running: bool = False,
        no_sched_changes: bool = False,
        err_on_sched_fail: bool = False,
    ):
        """
        :arg cache_running:
            Always cache running datastore data which makes mainly repeated retrieval of
            data much faster. Affects all sessions created on this connection.
        :arg no_sched_changes:
            Do not parse internal modules data and apply any scheduled changes. Makes
            creating the connection faster but, obviously, scheduled changes are not
            applied.
        :arg err_on_sched_fail:
            If applying any of the scheduled changes fails, do not create a connection
            and return an error.
        """
        flags = 0
        if cache_running:
            flags |= lib.SR_CONN_CACHE_RUNNING
        if no_sched_changes:
            flags |= lib.SR_CONN_NO_SCHED_CHANGES
        if err_on_sched_fail:
            flags |= lib.SR_CONN_ERR_ON_SCHED_FAIL
        conn_p = ffi.new("sr_conn_ctx_t **")
        check_call(lib.sr_connect, flags, conn_p)
        self.cdata = conn_p[0]

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.disconnect()

    def disconnect(self) -> None:
        """
        Disconnect from the sysrepo datastore.

        Cleans up and frees connection context allocated when instantiating the object.
        All sessions and subscriptions started within the connection will be
        automatically stopped and cleaned up too.

        Connection and all its associated sessions and subscriptions can no longer be
        used even on error.
        """
        try:
            check_call(lib.sr_disconnect, self.cdata)
        finally:
            self.cdata = None

    def start_session(self, datastore: str = "running") -> SysrepoSession:
        """
        Start a new session.

        :arg datastore:
            Datastore on which all sysrepo functions within this session will operate.
            Later on, datastore can be later changed using
            `SysrepoSession.switch_datastore`.

        :returns:
            A `SysrepoSession` object that can be used as a context manager. It will be
            automatically stopped when the manager exits::

                with conn.start_session() as sess:
                    # to stuff with sess
                # sess.stop() has been called whatever happens
        """
        ds = datastore_value(datastore)
        sess_p = ffi.new("sr_session_ctx_t **")
        check_call(lib.sr_session_start, self.cdata, ds, sess_p)
        return SysrepoSession(sess_p[0])

    def get_ly_ctx(self) -> libyang.Context:
        """
        :returns:
            The `libyang.Context` object associated with this connection.
        """
        ctx = lib.sr_get_context(self.cdata)
        if not ctx:
            raise SysrepoInternalError("sr_get_context failed")
        return libyang.Context(cdata=ctx)

    def install_module(
        self,
        filepath: str,
        searchdirs: Optional[str] = None,
        enabled_features: Sequence[str] = (),
    ) -> None:
        """
        Install a new schema (module) into sysrepo. Deferred until there are no
        connections!

        :arg filepath:
            Path to the new schema. Can have either YANG or YIN extension/format.
        :arg searchdirs:
            Optional search directories for import schemas, supports the format
            `<dir>[:<dir>]*`.
        :arg enabled_features:
            Array of enabled features.
        """
        if enabled_features:
            # convert to C strings array
            features = tuple(str2c(f) for f in enabled_features)
        else:
            features = ffi.NULL
        check_call(
            lib.sr_install_module,
            self.cdata,
            str2c(filepath),
            str2c(searchdirs),
            features,
            len(enabled_features),
        )

    def remove_module(self, name: str) -> None:
        """
        Remove an installed module from sysrepo. Deferred until there are no
        connections!

        :arg str name:
            Name of the module to remove.
        """
        check_call(lib.sr_remove_module, self.cdata, str2c(name))



# Copyright (c) 2020 6WIND S.A.
# SPDX-License-Identifier: BSD-3-Clause

import logging

from _sysrepo import ffi, lib
from .change import (
    Change,
    ChangeCreated,
    ChangeDeleted,
    ChangeModified,
    ChangeMoved,
    update_config_cache,
)
from .connection import SysrepoConnection
from .errors import (
    SysrepoCallbackFailedError,
    SysrepoCallbackShelveError,
    SysrepoError,
    SysrepoExistsError,
    SysrepoInternalError,
    SysrepoInvalArgError,
    SysrepoLockedError,
    SysrepoLyError,
    SysrepoNomemError,
    SysrepoNotFoundError,
    SysrepoOperationFailedError,
    SysrepoSysError,
    SysrepoTimeOutError,
    SysrepoUnauthorizedError,
    SysrepoUnsupportedError,
    SysrepoValidationFailedError,
)
from .util import c2str
from .value import (
    AnyData,
    AnyXML,
    Binary,
    Bits,
    Bool,
    Container,
    ContainerPresence,
    Decimal64,
    Enum,
    IdentityRef,
    InstanceId,
    Int8,
    Int16,
    Int32,
    Int64,
    LeafEmpty,
    List,
    String,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    Value,
)


__all__ = [
    "SysrepoConnection",
    "Change",
    "ChangeCreated",
    "ChangeDeleted",
    "ChangeModified",
    "ChangeMoved",
    "update_config_cache",
    "SysrepoError",
    "SysrepoCallbackFailedError",
    "SysrepoCallbackShelveError",
    "SysrepoExistsError",
    "SysrepoInternalError",
    "SysrepoInvalArgError",
    "SysrepoLockedError",
    "SysrepoLyError",
    "SysrepoNomemError",
    "SysrepoNotFoundError",
    "SysrepoOperationFailedError",
    "SysrepoSysError",
    "SysrepoTimeOutError",
    "SysrepoUnauthorizedError",
    "SysrepoUnsupportedError",
    "SysrepoValidationFailedError",
    "AnyData",
    "AnyXML",
    "Binary",
    "Bits",
    "Bool",
    "Container",
    "ContainerPresence",
    "Decimal64",
    "Enum",
    "IdentityRef",
    "InstanceId",
    "Int16",
    "Int32",
    "Int64",
    "Int8",
    "LeafEmpty",
    "List",
    "String",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt8",
    "Value",
]

LOG = logging.getLogger(__name__)
LOG.addHandler(logging.NullHandler())
lib.sr_log_set_cb(lib.srpy_log_cb)


@ffi.def_extern(name="srpy_log_cb")
def log_callback(level, msg):
    py_level = {
        lib.SR_LL_ERR: logging.ERROR,
        lib.SR_LL_WRN: logging.WARNING,
        lib.SR_LL_INF: logging.INFO,
        lib.SR_LL_DBG: logging.DEBUG,
    }.get(level, logging.NOTSET)
    LOG.log(py_level, "%s", c2str(msg))
